/* PostgreSQL Python Interface 
 * ---------------------------
 * Copyright (c) 2000 - Eric BIANCHI & Thierry MICHEL 
 * Copyright (c) 2000 - MixadLive srl Torino, Italy 
 * http://www.mixadlive.com 
 *
 * This library is free software, you can redistribute it and/or 
 * modify it under the terms of the GNU Lesser General Public Licence as 
 * published by the Free Software Foundation. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU Lesser General Public License for more details 
 * You should have received a copy of the GNU Lesser General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
 *
 * MixadLive, Srl., hereby disclaims all copyright
 * interest in the library `PoPy' written  
 * by Eric BIANCHI and Thierry MICHEL 
 *
 * -- Eric - Developer 
 * -- Thierry - Developer
 * For bugs & comments: eric@mixadlive.com, thierry@mixadlive.com 
 *
 * This PostgreSQL python interface is Python DB-API version 2.0
 * compliant. For more informations about DB-API specifications 
 * you can find them at : 
 *
 * http://www.python.org 
 *
 */

#include <Python.h>
#include <structmember.h>
#include <postgres.h>
#include <catalog/pg_type.h>
#include <libpq-fe.h>
#include <libpq/libpq-fs.h>
#include <regex.h>
#include <string.h>
#include <stdlib.h>

/* Handling Date and Time Objects */
#include <time.h>
#include "mx/mxDateTime.h"



static PyObject *PoPy_connect(PyObject *self,PyObject *args);
/* Must be cancelled */
#ifdef __DEBUG__
static PyObject *PoPy_test(PyObject *string,PyObject *dict);
#endif
static PyObject *PoPy__BinaryObject_new (Oid oid,PGconn *conx);

#define PG_LO_NOTDEFINED -1000
#define PG_FALSE 0
#define PG_TRUE  1
#define MAX_BUFFER 1024
#define PG_LOCREAT 1
#define PG_LOUPDATE 2

/* Column status */
#define PG_NULL REGPROCOID
#define PG_CHAR CHAROID
#define PG_VARCHAR VARCHAROID

/* PostgreSQL types definition */
#define PG_BOOLEAN BOOLOID
#define PG_FLOAT FLOAT4OID
#define PG_DOUBLE FLOAT8OID
#define PG_SHORT INT2OID
#define PG_INTEGER INT4OID
#define PG_DATE DATEOID
#define PG_TIME TIMEOID
#define PG_TIMESTAMP TIMESTAMPOID
#define PG_CASH CASHOID
#define PG_BINARY OIDOID
#define PG_ROWID 1000 /* Must be set bad value ! */


#define is_conx_object(ob) ((ob)->ob_type == &PoPy__ConnectionObject_Type)
#define Binary_Check(op) ((op)->ob_type == &PoPy__BinaryObject_Type)

#define SET_POPY_EXCEPTION(conn, res, type) do {  if(res == NULL){ \
                PyErr_SetString(PoPy_OperationalError,PQerrorMessage(conn));} \
                else {PyErr_SetString(type,PQresultErrorMessage(res));}\
                } while(0)

#ifdef __DEBUG__

#define DBG(s, a...) do { printf(s, ## a); printf("\n"); } while (0)

#define REF_CHECK(v) do { if ((v)->ob_refcnt != 1) { \
		printf("PoPy: refcnt bug on line %d, var %s, refs %d\n", \
			__LINE__, #v, (v)->ob_refcnt); \
		} } while (0)

#else

#define DBG(s, a...)
#define REF_CHECK(v)

#endif

static PyObject * PoPy_Error;

/********************************/
/* PG DB-API Objects definition */
/********************************/

/* Binary object */
typedef struct {
  PyObject_HEAD
  void *buffer;
  Oid oid;
  size_t size;
} PoPy__BinaryObject;

/* Cursor Object */
typedef struct {
  PyObject_HEAD
  int rowcount;
/*   int current; */
  char bool_style;
  int auto_commit;
  char cursor_name[15];
  int arraysize;
  char state;
  char executed;
  char req_type;
  PyObject *datestyle;
  PyObject *description;
  PGconn ** conn;
  PGresult *result;
} PoPy__CursorObject;

/* Connection Object */
typedef struct {
  PyObject_HEAD
  PGconn *conn;
  char bool_style;
  int open;
  int auto_commit;
  int commit_on;
/*   PyObject *cursor; */
  
} PoPy__ConnectionObject;

/* Exceptions definition */
static PyObject * PoPy_Warning;
static PyObject * PoPy_Error;
static PyObject * PoPy_InterfaceError;
static PyObject * PoPy_DatabaseError;
static PyObject * PoPy_DataError;
static PyObject * PoPy_OperationalError;
static PyObject * PoPy_IntegrityError;
static PyObject * PoPy_InternalError;
static PyObject * PoPy_ProgrammingError;
static PyObject * PoPy_NotSupportedError;


/* Types definition */
extern PyTypeObject PoPy__ConnectionObject_Type;
extern PyTypeObject PoPy__CursorObject_Type;
extern PyTypeObject PoPy__BinaryObject_Type;
